# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""EditorWidget and EditorMainWindow widgets."""

# pylint: disable=C0103
# pylint: disable=R0903
# pylint: disable=R0911
# pylint: disable=R0201

# Standard library imports
import logging
import os.path as osp
import sys

# Third party imports
import qstylizer.style
from qtpy.QtCore import QByteArray, QEvent, QPoint, QSize, Qt, Signal, Slot
from qtpy.QtGui import QFont
from qtpy.QtWidgets import (QAction, QApplication, QMainWindow, QSplitter,
                            QVBoxLayout, QWidget)

# Local imports
from spyder.api.plugins import Plugins
from spyder.api.config.decorators import on_conf_change
from spyder.api.config.mixins import SpyderConfigurationObserver
from spyder.api.widgets.toolbars import ApplicationToolbar
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.config.base import _
from spyder.plugins.editor.widgets.splitter import EditorSplitter
from spyder.plugins.editor.widgets.status import (CursorPositionStatus,
                                                  EncodingStatus, EOLStatus,
                                                  ReadWriteStatus, VCSStatus)
from spyder.plugins.mainmenu.api import (
    ApplicationMenu,
    ApplicationMenus,
    MENUBAR_STYLESHEET,
)
from spyder.plugins.outlineexplorer.main_widget import OutlineExplorerWidget
from spyder.plugins.toolbar.api import ApplicationToolbars
from spyder.py3compat import qbytearray_to_str
from spyder.utils.palette import SpyderPalette
from spyder.utils.qthelpers import create_toolbutton
from spyder.utils.stylesheet import APP_STYLESHEET
from spyder.widgets.findreplace import FindReplace


logger = logging.getLogger(__name__)


# ---- Constants
# -----------------------------------------------------------------------------
class EditorMainWindowMenus:
    View = "view"


class ViewMenuSections:
    Outline = "outline"
    Toolbars = "toolbars"


class EditorMainWindowActions:
    ToggleOutline = "toggle_outline"


# ---- Widgets
# -----------------------------------------------------------------------------
class OutlineExplorerInEditorWindow(OutlineExplorerWidget):

    sig_collapse_requested = Signal()

    @Slot()
    def close_dock(self):
        """
        Reimplemented to preserve the widget's visible state when shown in an
        editor window.
        """
        self.sig_collapse_requested.emit()


class EditorWidget(QSplitter, SpyderConfigurationObserver):
    """Main widget to show in EditorMainWindow."""

    CONF_SECTION = 'editor'
    SPLITTER_WIDTH = "7px"

    def __init__(self, parent, main_widget, menu_actions, outline_plugin):
        super().__init__(parent)
        self.setAttribute(Qt.WA_DeleteOnClose)

        # ---- Attributes
        self.editorstacks = []
        self.main_widget = main_widget
        self._sizes = None

        # This needs to be done at this point to avoid an error at startup
        self._splitter_css = self._generate_splitter_stylesheet()

        # ---- Find widget
        self.find_widget = FindReplace(self, enable_replace=True)
        self.find_widget.hide()

        # ---- Status bar
        statusbar = parent.statusBar()
        self.vcs_status = VCSStatus(self)
        self.cursorpos_status = CursorPositionStatus(self)
        self.encoding_status = EncodingStatus(self)
        self.eol_status = EOLStatus(self)
        self.readwrite_status = ReadWriteStatus(self)

        statusbar.insertPermanentWidget(0, self.readwrite_status)
        statusbar.insertPermanentWidget(0, self.eol_status)
        statusbar.insertPermanentWidget(0, self.encoding_status)
        statusbar.insertPermanentWidget(0, self.cursorpos_status)
        statusbar.insertPermanentWidget(0, self.vcs_status)

        # ---- Outline.
        self.outlineexplorer = None
        if outline_plugin is not None:
            self.outlineexplorer = OutlineExplorerInEditorWindow(
                'outline_explorer',
                outline_plugin,
                self,
                context=f'editor_window_{str(id(self))}'
            )

            # Show widget's toolbar
            self.outlineexplorer.setup()
            self.outlineexplorer.update_actions()
            self.outlineexplorer._setup()
            self.outlineexplorer.render_toolbars()

            # Remove bottom section actions from Options menu because they
            # don't apply here.
            options_menu = self.outlineexplorer.get_options_menu()
            for action in ['undock_pane', 'lock_unlock_position']:
                options_menu.remove_action(action)

            # Signals
            self.outlineexplorer.edit_goto.connect(
                lambda filenames, goto, word:
                main_widget.load(filenames=filenames, goto=goto, word=word,
                                 editorwindow=self.parent())
            )

            self.outlineexplorer.sig_collapse_requested.connect(
                lambda: self.set_conf("show_outline_in_editor_window", False)
            )

            # Start symbol services for all supported languages
            for language in outline_plugin.get_supported_languages():
                self.outlineexplorer.start_symbol_services(language)

            # Tell Outline's treewidget that is visible
            self.outlineexplorer.change_tree_visibility(True)

        # ---- Editor widgets
        editor_widgets = QWidget(self)
        editor_layout = QVBoxLayout()
        editor_layout.setSpacing(0)
        editor_layout.setContentsMargins(0, 0, 0, 0)
        editor_widgets.setLayout(editor_layout)
        self.editorsplitter = EditorSplitter(
            self,
            main_widget,
            menu_actions,
            register_editorstack_cb=self.register_editorstack,
            unregister_editorstack_cb=self.unregister_editorstack
        )
        editor_layout.addWidget(self.editorsplitter)
        editor_layout.addWidget(self.find_widget)

        # ---- Splitter
        self.splitter = QSplitter(self)
        self.splitter.setContentsMargins(0, 0, 0, 0)
        if self.outlineexplorer is not None:
            self.splitter.addWidget(self.outlineexplorer)
        self.splitter.addWidget(editor_widgets)

        self.splitter.setStretchFactor(0, 1)
        self.splitter.setStretchFactor(1, 4)

        # This sets the same UX as the one users encounter when the editor is
        # maximized.
        self.splitter.setChildrenCollapsible(False)

        self.splitter.splitterMoved.connect(self.on_splitter_moved)

        if (
            self.outlineexplorer is not None
            and not self.get_conf("show_outline_in_editor_window")
        ):
            self.outlineexplorer.close_dock()

        # ---- Style
        self.splitter.setStyleSheet(self._splitter_css.toString())

    def register_editorstack(self, editorstack):
        logger.debug("Registering editorstack")
        self.__print_editorstacks()

        self.editorstacks.append(editorstack)
        self.main_widget.last_focused_editorstack[self.parent()] = editorstack
        editorstack.set_closable(len(self.editorstacks) > 1)
        editorstack.set_outlineexplorer(self.outlineexplorer)
        editorstack.set_find_widget(self.find_widget)
        editorstack.reset_statusbar.connect(self.readwrite_status.hide)
        editorstack.reset_statusbar.connect(self.encoding_status.hide)
        editorstack.reset_statusbar.connect(self.cursorpos_status.hide)
        editorstack.readonly_changed.connect(
            self.readwrite_status.update_readonly)
        editorstack.encoding_changed.connect(
            self.encoding_status.update_encoding)
        editorstack.sig_editor_cursor_position_changed.connect(
            self.cursorpos_status.update_cursor_position)
        editorstack.sig_refresh_eol_chars.connect(self.eol_status.update_eol)
        self.main_widget.register_editorstack(editorstack)

    def __print_editorstacks(self):
        logger.debug(
            f"{len(self.editorstacks)} editorstack(s) in EditorWidget:"
        )
        for es in self.editorstacks:
            logger.debug(f"    {es}")

    def _generate_splitter_stylesheet(self):
        # Set background color to be the same as the one used in any other
        # widget. This removes what appears to be some extra borders in several
        # places.
        css = qstylizer.style.StyleSheet()
        css.QSplitter.setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_1
        )

        # Make splitter handle to have the same size as the QMainWindow
        # separators. That's because the editor and outline are shown like
        # this when the editor is maximized.
        css['QSplitter::handle'].setValues(
            width=self.SPLITTER_WIDTH,
            height=self.SPLITTER_WIDTH
        )

        return css

    def unregister_editorstack(self, editorstack):
        logger.debug("Unregistering editorstack")
        self.main_widget.unregister_editorstack(editorstack)
        self.editorstacks.pop(self.editorstacks.index(editorstack))
        self.__print_editorstacks()

    def unregister_all_editorstacks(self):
        logger.debug("Unregistering all editorstacks")
        for es in self.editorstacks:
            es.close()

    @Slot(object)
    def on_window_state_changed(self, window_state):
        """
        Actions to take when the parent window state has changed.
        """
        # There's no need to update the Outline when the window is minimized
        if window_state == Qt.WindowMinimized:
            self.outlineexplorer.change_tree_visibility(False)
        else:
            self.outlineexplorer.change_tree_visibility(True)

    def on_splitter_moved(self, position, index):
        """Actions to take when the splitter is moved."""
        # There's no need to update the Outline when the user moves the
        # splitter to hide it.
        # Note: The 20 below was selected because the Outline can't have that
        # small width. So, if the splitter position plus that amount is greater
        # than the total widget width, it means the Outline was collapsed.
        if (position + 20) > self.size().width():
            self.outlineexplorer.change_tree_visibility(False)
        else:
            self.outlineexplorer.change_tree_visibility(True)

    @on_conf_change(option='show_outline_in_editor_window')
    def toggle_outlineexplorer(self, value):
        """Toggle outline explorer visibility."""
        if value:
            # When self._sizes is not available, the splitter sizes are set
            # automatically by the ratios set for it above.
            if self._sizes is not None:
                self.splitter.setSizes(self._sizes)

            # Show and enable splitter handle
            self._splitter_css['QSplitter::handle'].setValues(
                width=self.SPLITTER_WIDTH,
                height=self.SPLITTER_WIDTH
            )
            self.splitter.setStyleSheet(self._splitter_css.toString())
            if self.splitter.handle(1) is not None:
                self.splitter.handle(1).setEnabled(True)
        else:
            self._sizes = self.splitter.sizes()
            self.splitter.setChildrenCollapsible(True)

            # Collapse Outline
            self.splitter.moveSplitter(0, 1)

            # Hide and disable splitter handle
            self._splitter_css['QSplitter::handle'].setValues(
                width="0px",
                height="0px"
            )
            self.splitter.setStyleSheet(self._splitter_css.toString())
            if self.splitter.handle(1) is not None:
                self.splitter.handle(1).setEnabled(False)

            self.splitter.setChildrenCollapsible(False)


class EditorMainWindow(QMainWindow, SpyderWidgetMixin):
    CONF_SECTION = "editor"

    sig_window_state_changed = Signal(object)

    def __init__(self, main_widget, menu_actions, outline_plugin, parent=None):
        # Parent needs to be `None` if the created widget is meant to be
        # independent. See spyder-ide/spyder#17803
        super().__init__(parent, class_parent=main_widget)
        self.setAttribute(Qt.WA_DeleteOnClose)

        # ---- Attributes
        self.main_widget = main_widget
        self.window_size = None
        self.toolbars = []

        # ---- Main widget
        self.editorwidget = EditorWidget(
            self,
            main_widget,
            menu_actions,
            outline_plugin
        )
        self.sig_window_state_changed.connect(
            self.editorwidget.on_window_state_changed
        )
        self.setCentralWidget(self.editorwidget)

        # ---- Style
        self.setStyleSheet(str(APP_STYLESHEET))
        if not sys.platform == "darwin":
            self.menuBar().setStyleSheet(str(MENUBAR_STYLESHEET))

        # Give focus to current editor to update/show all status bar widgets
        editorstack = self.editorwidget.editorsplitter.editorstack
        editor = editorstack.get_current_editor()
        if editor is not None:
            editor.setFocus()

        self.setWindowTitle("Spyder - %s" % main_widget.windowTitle())
        self.setWindowIcon(main_widget.windowIcon())

        # ---- Add toolbars
        toolbar_list = [
            ApplicationToolbars.File,
            ApplicationToolbars.Run,
            ApplicationToolbars.Debug
        ]

        for toolbar_id in toolbar_list:
            # This is necessary to run tests for this widget without Spyder's
            # main window
            try:
                toolbar = self.get_toolbar(toolbar_id, plugin=Plugins.Toolbar)
            except KeyError:
                continue

            new_toolbar = ApplicationToolbar(self, toolbar_id, toolbar._title)
            for action in toolbar.actions():
                new_toolbar.add_item(action)

            new_toolbar.render()
            new_toolbar.setMovable(False)

            self.addToolBar(new_toolbar)
            self.toolbars.append(new_toolbar)

        # ---- Add menus
        menu_list = [
            ApplicationMenus.File,
            ApplicationMenus.Edit,
            ApplicationMenus.Search,
            ApplicationMenus.Source,
            ApplicationMenus.Run,
            ApplicationMenus.Tools,
            EditorMainWindowMenus.View,
            ApplicationMenus.Help
        ]

        for menu_id in menu_list:
            if menu_id == EditorMainWindowMenus.View:
                view_menu = self._create_view_menu()
                self.menuBar().addMenu(view_menu)
            else:
                # This is necessary to run tests for this widget without
                # Spyder's main window
                try:
                    self.menuBar().addMenu(
                        self.get_menu(menu_id, plugin=Plugins.MainMenu)
                    )
                except KeyError:
                    continue

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def resizeEvent(self, event):
        """Reimplement Qt method"""
        if not self.isMaximized() and not self.isFullScreen():
            self.window_size = self.size()
        QMainWindow.resizeEvent(self, event)

    def closeEvent(self, event):
        """Reimplement Qt method"""
        self.editorwidget.unregister_all_editorstacks()
        if self.main_widget.windowwidget is not None:
            self.main_widget.dockwidget.setWidget(self.main_widget)
            self.main_widget.dockwidget.setVisible(True)
        self.main_widget.switch_to_plugin()
        QMainWindow.closeEvent(self, event)
        if self.main_widget.windowwidget is not None:
            self.main_widget.windowwidget = None

    def changeEvent(self, event):
        """
        Override Qt method to emit a custom `sig_windowstate_changed` signal
        when there's a change in the window state.
        """
        if event.type() == QEvent.WindowStateChange:
            self.sig_window_state_changed.emit(self.windowState())
        super().changeEvent(event)

    # ---- Public API
    # -------------------------------------------------------------------------
    def get_layout_settings(self):
        """Return layout state"""
        splitsettings = self.editorwidget.editorsplitter.get_layout_settings()
        return dict(size=(self.window_size.width(), self.window_size.height()),
                    pos=(self.pos().x(), self.pos().y()),
                    is_maximized=self.isMaximized(),
                    is_fullscreen=self.isFullScreen(),
                    hexstate=qbytearray_to_str(self.saveState()),
                    splitsettings=splitsettings)

    def set_layout_settings(self, settings):
        """Restore layout state"""
        size = settings.get('size')
        if size is not None:
            self.resize(QSize(*size))
            self.window_size = self.size()
        pos = settings.get('pos')
        if pos is not None:
            self.move(QPoint(*pos))
        hexstate = settings.get('hexstate')
        if hexstate is not None:
            self.restoreState(
                QByteArray().fromHex(str(hexstate).encode('utf-8'))
            )
        if settings.get('is_maximized'):
            self.setWindowState(Qt.WindowMaximized)
        if settings.get('is_fullscreen'):
            self.setWindowState(Qt.WindowFullScreen)
        splitsettings = settings.get('splitsettings')
        if splitsettings is not None:
            self.editorwidget.editorsplitter.set_layout_settings(splitsettings)

    # ---- Private API
    # -------------------------------------------------------------------------
    def _create_view_menu(self):
        # Create menu
        view_menu = self._create_menu(
            menu_id=EditorMainWindowMenus.View,
            parent=self,
            title=_("&View"),
            register=False,
            MenuClass=ApplicationMenu
        )

        # Create Outline action
        self.toggle_outline_action = self.create_action(
            EditorMainWindowActions.ToggleOutline,
            _("Outline"),
            toggled=True,
            option="show_outline_in_editor_window"
        )

        view_menu.add_action(
            self.toggle_outline_action,
            section=ViewMenuSections.Outline
        )

        # Add toolbar toggle view actions
        visible_toolbars = self.get_conf(
            'last_visible_toolbars',
            section='toolbar'
        )

        for toolbar in self.toolbars:
            toolbar_action = toolbar.toggleViewAction()
            toolbar_action.action_id = f'toolbar_{toolbar.ID}'

            if toolbar.ID not in visible_toolbars:
                toolbar_action.setChecked(False)
                toolbar.setVisible(False)
            else:
                toolbar_action.setChecked(True)
                toolbar.setVisible(True)

            view_menu.add_action(
                toolbar_action,
                section=ViewMenuSections.Toolbars
            )

        return view_menu


class EditorMainWidgetExample(QSplitter):
    def __init__(self):
        QSplitter.__init__(self)

        self._plugin = None

        self.dock_action = None
        self.undock_action = None
        self.close_action = None
        self.windowwidget = None
        self.lock_unlock_action = None
        menu_actions = []

        self.editorstacks = []
        self.editorwindows = []

        self.last_focused_editorstack = {}  # fake

        self.find_widget = FindReplace(self, enable_replace=True)
        self.outlineexplorer = OutlineExplorerWidget(None, self, self)
        self.outlineexplorer.edit_goto.connect(self.go_to_file)
        self.editor_splitter = EditorSplitter(self, self, menu_actions,
                                              first=True,
                                              use_switcher=False)

        editor_widgets = QWidget(self)
        editor_layout = QVBoxLayout()
        editor_layout.setSpacing(0)
        editor_layout.setContentsMargins(0, 0, 0, 0)
        editor_widgets.setLayout(editor_layout)
        editor_layout.addWidget(self.editor_splitter)
        editor_layout.addWidget(self.find_widget)

        self.setContentsMargins(0, 0, 0, 0)
        self.addWidget(editor_widgets)
        self.addWidget(self.outlineexplorer)

        self.setStretchFactor(0, 5)
        self.setStretchFactor(1, 1)

        self.menu_actions = menu_actions
        self.toolbar_list = None
        self.menu_list = None
        self.setup_window([], [])

    def go_to_file(self, fname, lineno, text='', start_column=None):
        editorstack = self.editorstacks[0]
        editorstack.set_current_filename(str(fname))
        editor = editorstack.get_current_editor()
        editor.go_to_line(lineno, word=text, start_column=start_column)

    def closeEvent(self, event):
        for win in self.editorwindows[:]:
            win.close()
        logger.debug("%d: %r" % (len(self.editorwindows), self.editorwindows))
        logger.debug("%d: %r" % (len(self.editorstacks), self.editorstacks))
        event.accept()

    def load(self, fname):
        QApplication.processEvents()
        editorstack = self.editorstacks[0]
        editorstack.load(fname)
        editorstack.analyze_script()

    def register_editorstack(self, editorstack):
        logger.debug(
            "FakeEditorMainWidget.register_editorstack: %r" % editorstack
        )
        self.editorstacks.append(editorstack)
        if self.isAncestorOf(editorstack):
            # editorstack is a child of the EditorMainWidget
            editorstack.set_closable(len(self.editorstacks) > 1)
            editorstack.set_outlineexplorer(self.outlineexplorer)
            editorstack.set_find_widget(self.find_widget)
            oe_btn = create_toolbutton(self)
            editorstack.add_corner_widgets_to_tabbar([5, oe_btn])

        action = QAction(self)
        editorstack.set_io_actions(action, action, action, action)
        font = QFont("Courier New")
        font.setPointSize(10)
        editorstack.set_default_font(font, color_scheme='Spyder')

        editorstack.sig_close_file.connect(self.close_file_in_all_editorstacks)
        editorstack.file_saved.connect(self.file_saved_in_editorstack)
        editorstack.file_renamed_in_data.connect(
                                      self.file_renamed_in_data_in_editorstack)
        editorstack.plugin_load.connect(self.load)

    def unregister_editorstack(self, editorstack):
        logger.debug(
            "EditorMainWidget.unregister_editorstack: %r" % editorstack
        )
        self.editorstacks.pop(self.editorstacks.index(editorstack))

    def clone_editorstack(self, editorstack):
        editorstack.clone_from(self.editorstacks[0])

    def setup_window(self, toolbar_list, menu_list):
        self.toolbar_list = toolbar_list
        self.menu_list = menu_list

    def create_new_window(self):
        window = EditorMainWindow(self, self.menu_actions,
                                  self.toolbar_list, self.menu_list,
                                  show_fullpath=False, show_all_files=False,
                                  group_cells=True, show_comments=True,
                                  sort_files_alphabetically=False)
        window.resize(self.size())
        window.show()
        self.register_editorwindow(window)
        window.destroyed.connect(lambda: self.unregister_editorwindow(window))

    def register_editorwindow(self, window):
        logger.debug("register_editorwindowQObject*: %r" % window)
        self.editorwindows.append(window)

    def unregister_editorwindow(self, window):
        logger.debug("unregister_editorwindow: %r" % window)
        self.editorwindows.pop(self.editorwindows.index(window))

    def get_focus_widget(self):
        pass

    @Slot(str, str)
    def close_file_in_all_editorstacks(self, editorstack_id_str, filename):
        for editorstack in self.editorstacks:
            if str(id(editorstack)) != editorstack_id_str:
                editorstack.blockSignals(True)
                index = editorstack.get_index_from_filename(filename)
                editorstack.close_file(index, force=True)
                editorstack.blockSignals(False)

    # This method is never called in this plugin example. It's here only
    # to show how to use the file_saved signal (see above).
    @Slot(str, str, str)
    def file_saved_in_editorstack(self, editorstack_id_str,
                                  original_filename, filename):
        """A file was saved in editorstack, this notifies others"""
        for editorstack in self.editorstacks:
            if str(id(editorstack)) != editorstack_id_str:
                editorstack.file_saved_in_other_editorstack(original_filename,
                                                            filename)

    # This method is never called in this plugin example. It's here only
    # to show how to use the file_saved signal (see above).
    @Slot(str, str, str)
    def file_renamed_in_data_in_editorstack(
        self, original_filename, filename, editorstack_id_str
    ):
        """A file was renamed in data in editorstack, this notifies others"""
        for editorstack in self.editorstacks:
            if str(id(editorstack)) != editorstack_id_str:
                editorstack.rename_in_data(original_filename, filename)

    def _get_color_scheme(self):
        pass


def test():
    from spyder.utils.qthelpers import qapplication
    from spyder.config.base import get_module_path

    spyder_dir = get_module_path('spyder')
    app = qapplication(test_time=8)

    test = EditorMainWidgetExample()
    test.resize(900, 700)
    test.show()

    import time
    t0 = time.time()
    test.load(osp.join(spyder_dir, "widgets", "collectionseditor.py"))
    test.load(osp.join(spyder_dir, "plugins", "editor", "widgets",
                       "window.py"))
    test.load(osp.join(spyder_dir, "plugins", "explorer", "widgets",
                       'explorer.py'))
    test.load(osp.join(spyder_dir, "plugins", "editor", "widgets",
                       "codeeditor", "codeeditor.py"))
    print("Elapsed time: %.3f s" % (time.time()-t0))  # spyder: test-skip

    sys.exit(app.exec_())


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests syncing between the EditorStack and OutlineExplorerWidget.
"""

# Standard library imports
import os
import os.path as osp
import sys
from unittest.mock import Mock

# Third party imports
import pytest

# Local imports
from spyder.config.base import running_in_ci
from spyder.plugins.editor.widgets.editorstack import EditorStack


def get_tree_elements(treewidget):
    """Get elements present in the Outline tree widget."""
    root_item = treewidget.get_top_level_items()[0]
    root_ref = root_item.ref
    filename = osp.basename(root_ref.name)
    root_tree = {filename: []}
    stack = [(root_tree[filename], node) for node in root_ref.children]

    while len(stack) > 0:
        parent_tree, node = stack.pop(0)
        this_tree = {node.name: []}
        parent_tree.append(this_tree)
        this_stack = [(this_tree[node.name], child)
                      for child in node.children]
        stack = this_stack + stack
    return root_tree


# ---- Qt Test Fixtures
@pytest.fixture(scope="module")
def test_files(tmpdir_factory):
    """Create and save some python codes and text in temporary files."""
    tmpdir = tmpdir_factory.mktemp("files")

    filename1 = osp.join(tmpdir.strpath, 'foo1.py')
    with open(filename1, 'w') as f:
        f.write("# -*- coding: utf-8 -*-\n"
                "def foo:\n"
                "    print(Hello World!)\n")

    filename2 = osp.join(tmpdir.strpath, 'text1.txt')
    with open(filename2, 'w') as f:
        f.write("This is a simple text file for\n"
                "testing the Outline Explorer.\n")

    filename3 = osp.join(tmpdir.strpath, 'foo2.py')
    with open(filename3, 'w') as f:
        f.write("# -*- coding: utf-8 -*-\n"
                "# ---- a comment\n")

    return [filename1, filename2, filename3]


@pytest.fixture
def editorstack(qtbot, outlineexplorer):
    def _create_editorstack(files):
        editorstack = EditorStack(None, [], False)
        editorstack.set_find_widget(Mock())
        editorstack.set_io_actions(Mock(), Mock(), Mock(), Mock())
        editorstack.analysis_timer = Mock()
        editorstack.save_dialog_on_tests = True
        editorstack.set_outlineexplorer(outlineexplorer)

        qtbot.addWidget(editorstack)
        editorstack.show()

        for index, file in enumerate(files):
            focus = index == 0
            editorstack.load(file, set_current=focus)
        return editorstack
    return _create_editorstack


# ---- Test all files mode
def test_load_files(editorstack, outlineexplorer, test_files):
    """
    Test that the content of the outline explorer is updated correctly
    after a file is loaded in the editor.
    """
    editorstack = editorstack([])
    treewidget = outlineexplorer.treewidget

    # Load the test files one by one and assert the content of the
    # outline explorer.
    expected_result = [['foo1.py'],
                       ['foo1.py', 'text1.txt'],
                       ['foo1.py', 'text1.txt', 'foo2.py']]
    for index, file in enumerate(test_files):
        editorstack.load(file)
        assert editorstack.get_current_filename() == file
        assert editorstack.get_stack_index() == index

        results = [item.text(0) for item in treewidget.get_visible_items()]
        assert results == expected_result[index]
        assert editorstack.get_stack_index() == index


def test_close_editor(editorstack, outlineexplorer, test_files):
    """
    Test that the content of the outline explorer is empty after the
    editorstack has been closed.

    Regression test for spyder-ide/spyder#7798.
    """
    editorstack = editorstack(test_files)
    treewidget = outlineexplorer.treewidget
    assert treewidget.get_visible_items()

    # Close the editor and assert that the outline explorer tree is empty.
    editorstack.close()
    assert not treewidget.get_visible_items()


def test_close_a_file(editorstack, outlineexplorer, test_files):
    """
    Test that the content of the outline explorer is updated corrrectly
    after a file has been closed in the editorstack.

    Regression test for spyder-ide/spyder#7798.
    """
    editorstack = editorstack(test_files)
    treewidget = outlineexplorer.treewidget

    # Close 'foo2.py' and assert that the content of the outline explorer
    # tree has been updated.
    editorstack.close_file(index=1)
    results = [item.text(0) for item in treewidget.get_visible_items()]
    assert results == ['foo1.py', 'foo2.py']


def test_sort_file_alphabetically(editorstack, outlineexplorer, test_files):
    """
    Test that the option to sort the files in alphabetical order in the
    outline explorer is working as expected.

    This feature was introduced in spyder-ide/spyder#8015.
    """
    editorstack = editorstack(test_files)
    treewidget = outlineexplorer.treewidget
    results = [item.text(0) for item in treewidget.get_visible_items()]
    assert results == ['foo1.py', 'text1.txt', 'foo2.py']

    # Set the option to sort files alphabetically to True.
    treewidget.toggle_sort_files_alphabetically(True)
    results = [item.text(0) for item in treewidget.get_visible_items()]
    assert results == ['foo1.py', 'foo2.py', 'text1.txt']


def test_sync_file_order(editorstack, outlineexplorer, test_files):
    """
    Test that the order of the files in the Outline Explorer is updated when
    tabs are moved in the EditorStack.

    This feature was introduced in spyder-ide/spyder#8015.
    """
    editorstack = editorstack(test_files)
    treewidget = outlineexplorer.treewidget
    results = [item.text(0) for item in treewidget.get_visible_items()]
    assert results == ['foo1.py', 'text1.txt', 'foo2.py']

    # Switch tab 1 with tab 2.
    editorstack.tabs.tabBar().moveTab(0, 1)
    results = [item.text(0) for item in treewidget.get_visible_items()]
    assert results == ['text1.txt', 'foo1.py', 'foo2.py']


# ---- Test single file mode
@pytest.mark.skipif(
    running_in_ci() or os.name == 'nt',
    reason="Fails on CIs and on Windows"
)
def test_toggle_off_show_all_files(editorstack, outlineexplorer, test_files,
                                   qtbot):
    """
    Test that toggling off the option to show all files in the Outline Explorer
    hide all root file items but the one corresponding to the currently
    selected Editor and assert that the remaning root file item is
    expanded correctly.
    """
    editorstack = editorstack(test_files)
    treewidget = outlineexplorer.treewidget
    assert editorstack.get_stack_index() == 0

    # Untoggle show all files option.
    treewidget.toggle_show_all_files(False)
    qtbot.wait(500)
    results = [item.text(0) for item in treewidget.get_visible_items()]
    assert results == ['foo1.py']


@pytest.mark.skipif(sys.platform.startswith('linux'), reason="Fails on Linux")
def test_single_file_sync(editorstack, outlineexplorer, test_files, qtbot):
    """
    Test that the content of the Outline Explorer is updated correctly
    when the current Editor in the Editorstack changes.
    """
    editorstack = editorstack(test_files)
    treewidget = outlineexplorer.treewidget
    treewidget.toggle_show_all_files(False)
    assert editorstack.get_stack_index() == 0

    # Select the last file in the Editorstack.
    with qtbot.waitSignal(editorstack.editor_focus_changed):
        editorstack.tabs.setCurrentIndex(2)
    results = [item.text(0) for item in treewidget.get_visible_items()]
    assert results == ['foo2.py']


def test_toggle_on_show_all_files(editorstack, outlineexplorer, test_files):
    """
    Test that toggling back the option to show all files, after the
    order of the files in the Editorstack was changed while it was in single
    file mode, show all the root file items in the correct order.
    """
    editorstack = editorstack(test_files)
    treewidget = outlineexplorer.treewidget
    treewidget.toggle_show_all_files(False)

    # Move the first file to the second position in the tabbar of the
    # Editorstack and toggle back the show all files option.
    editorstack.tabs.tabBar().moveTab(0, 1)
    treewidget.toggle_show_all_files(True)
    results = [item.text(0) for item in treewidget.get_visible_items()]
    assert results == ['text1.txt', 'foo1.py', 'foo2.py']


if __name__ == "__main__":
    pytest.main([os.path.basename(__file__), '-vv', '-rw'])



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
In addition to the remote_call mechanism implemented in CommBase:
 - Send a message to a debugging kernel
"""
from contextlib import contextmanager
import logging

from qtpy.QtCore import QEventLoop, QObject, QTimer, Signal

from spyder_kernels.comms.commbase import CommBase, CommsErrorWrapper

from spyder.config.base import (
    get_debug_level, running_under_pytest)

logger = logging.getLogger(__name__)
TIMEOUT_KERNEL_START = 30


class KernelComm(CommBase, QObject):
    """
    Class with the necessary attributes and methods to handle
    communications with a console.
    """

    _sig_got_reply = Signal()
    sig_exception_occurred = Signal(dict)
    sig_comm_ready = Signal()

    def __init__(self):
        super(KernelComm, self).__init__()
        self.kernel_client = None

        # Register handlers
        self.register_call_handler('_async_error', self._async_error)
        self.register_call_handler('_comm_ready', self._comm_ready)

    def is_open(self, comm_id=None):
        """
        Check to see if the comm is open and ready to communicate.
        """
        id_list = self.get_comm_id_list(comm_id)
        if len(id_list) == 0:
            return False
        return all([self._comms[cid]['status'] == 'ready' for cid in id_list])

    @contextmanager
    def comm_channel_manager(self, comm_id, queue_message=False):
        """Use control_channel instead of shell_channel."""
        if queue_message:
            # Send without control_channel
            yield
            return

        id_list = self.get_comm_id_list(comm_id)
        for comm_id in id_list:
            self._comms[comm_id]['comm']._send_channel = (
                self.kernel_client.control_channel)
        try:
            yield
        finally:
            id_list = self.get_comm_id_list(comm_id)
            for comm_id in id_list:
                self._comms[comm_id]['comm']._send_channel = (
                    self.kernel_client.shell_channel)

    def _set_call_return_value(self, call_dict, return_value, is_error=False):
        """Override to use the comm_channel for all replies."""
        with self.comm_channel_manager(self.calling_comm_id, False):
            if is_error and (get_debug_level() or running_under_pytest()):
                # Disable error muting when debugging or testing
                call_dict['settings']['display_error'] = True
            super(KernelComm, self)._set_call_return_value(
                call_dict, return_value, is_error=is_error
            )

    def remove(self, comm_id=None, only_closing=False):
        """
        Remove the comm without notifying the other side.

        Use when the other side is already down.
        """
        id_list = self.get_comm_id_list(comm_id)
        for comm_id in id_list:
            if only_closing and self._comms[comm_id]['status'] != 'closing':
                continue
            del self._comms[comm_id]

    def close(self, comm_id=None):
        """Ask kernel to close comm and send confirmation."""
        id_list = self.get_comm_id_list(comm_id)
        for comm_id in id_list:
            # Send comm_close directly to avoid really closing the comm
            self._comms[comm_id]['comm']._send_msg(
                'comm_close', {}, None, None, None)
            self._comms[comm_id]['status'] = 'closing'

    def open_comm(self, kernel_client):
        """Open comm through the kernel client."""
        self.kernel_client = kernel_client
        try:
            logger.debug(
                f"Opening kernel comm for "
                f"{'<' + repr(kernel_client).split('.')[-1]}"
            )

            self._register_comm(
                # Create new comm and send the highest protocol
                kernel_client.comm_manager.new_comm(self._comm_name)
            )
        except AttributeError:
            logger.info(
                "Unable to open comm due to unexistent comm manager: " +
                "kernel_client.comm_manager=" + str(kernel_client.comm_manager)
            )

    def remote_call(self, interrupt=False, blocking=False, callback=None,
                    comm_id=None, timeout=None, display_error=False):
        """Get a handler for remote calls."""
        return super(KernelComm, self).remote_call(
            interrupt=interrupt, blocking=blocking, callback=callback,
            comm_id=comm_id, timeout=timeout, display_error=display_error)

    def on_incoming_call(self, call_dict):
        """A call was received"""
        super().on_incoming_call(call_dict)
        # Just in case the call was not received
        self._comm_ready()

    # ---- Private -----
    def _comm_ready(self):
        """If this function is called, the comm is ready"""
        if self._comms[self.calling_comm_id]['status'] != 'ready':
            self._comms[self.calling_comm_id]['status'] = 'ready'
            self.sig_comm_ready.emit()

    def _send_call(self, call_dict, comm_id, buffers):
        """Send call and interupt the kernel if needed."""
        settings = call_dict['settings']
        blocking = 'blocking' in settings and settings['blocking']
        interrupt = 'interrupt' in settings and settings['interrupt']
        queue_message = not interrupt and not blocking

        if not self.kernel_client.is_alive():
            if blocking:
                raise RuntimeError("Kernel is dead")
            else:
                # The user has other problems
                logger.info(
                    "Dropping message because kernel is dead: %s",
                    str(call_dict)
                )
                return

        with self.comm_channel_manager(
                comm_id, queue_message=queue_message):
            return super(KernelComm, self)._send_call(
                call_dict, comm_id, buffers
            )

    def _get_call_return_value(self, call_dict, comm_id):
        """
        Catch exception if call is not blocking.
        """
        try:
            return super(KernelComm, self)._get_call_return_value(
                call_dict, comm_id)
        except RuntimeError as e:
            settings = call_dict['settings']
            blocking = 'blocking' in settings and settings['blocking']
            if blocking:
                raise
            else:
                # The user has other problems
                logger.info(
                    "Dropping message because of exception: ",
                    str(e),
                    str(call_dict)
                )
                return

    def _wait_reply(self, comm_id, call_id, call_name, timeout):
        """Wait for the other side reply."""

        def got_reply():
            return call_id in self._reply_inbox

        timeout_msg = "Timeout while waiting for {}".format(
            self._reply_waitlist)
        self._wait(got_reply, self._sig_got_reply, timeout_msg, timeout)

    def _wait(self, condition, signal, timeout_msg, timeout):
        """
        Wait until condition() is True by running an event loop.

        signal: qt signal that should interrupt the event loop.
        timeout_msg: Message to display in case of a timeout.
        timeout: time in seconds before a timeout
        """
        # Exit if condition is fulfilled or the kernel is dead.
        if condition():
            return
        if not self.kernel_client.is_alive():
            raise RuntimeError("Kernel is dead")

        # Create event loop to wait with
        wait_loop = QEventLoop(None)
        wait_timeout = QTimer(self)
        wait_timeout.setSingleShot(True)

        # Connect signals to stop kernel loop
        wait_timeout.timeout.connect(wait_loop.quit)
        self.kernel_client.hb_channel.kernel_died.connect(wait_loop.quit)
        signal.connect(wait_loop.quit)

        # Wait until the kernel returns the value
        wait_timeout.start(timeout * 1000)
        while not condition():
            if not wait_timeout.isActive():
                signal.disconnect(wait_loop.quit)
                self.kernel_client.hb_channel.kernel_died.disconnect(
                    wait_loop.quit)
                if condition():
                    return
                if not self.kernel_client.is_alive():
                    raise RuntimeError("Kernel is dead")
                raise TimeoutError(timeout_msg)
            wait_loop.exec_()

        wait_timeout.stop()
        signal.disconnect(wait_loop.quit)
        self.kernel_client.hb_channel.kernel_died.disconnect(
            wait_loop.quit)

    def _handle_remote_call_reply(self, *args, **kwargs):
        """
        A blocking call received a reply.
        """
        super(KernelComm, self)._handle_remote_call_reply(*args, **kwargs)
        self._sig_got_reply.emit()

    def _async_error(self, error_wrapper):
        """
        Handle an error that was raised on the other side and sent back.
        """
        error_wrapper = CommsErrorWrapper.from_json(error_wrapper)
        for line in error_wrapper.format_error():
            self.sig_exception_occurred.emit(
                dict(text=line, is_traceback=True)
            )



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2013-2016 Colin Duquesnoy and others (see pyqode/AUTHORS.rst)
# Copyright (c) 2016- Spyder Project Contributors (see AUTHORS.txt)
#
# Distributed under the terms of the MIT License
# (see NOTICE.txt in the Spyder root directory for details)
# -----------------------------------------------------------------------------

"""
This module contains the marker panel.

Adapted from pyqode/core/panels/folding.py of the PyQode project
https://github.com/pyQode/pyQode

Original file:
https://github.com/pyQode/pyqode.core/blob/master/pyqode/core/panels/folding.py
"""

# Standard library imports
from math import ceil

# Third party imports
from intervaltree import IntervalTree
from qtpy.QtCore import Signal, QSize, QPointF, QRectF, QRect, Qt
from qtpy.QtGui import (QTextBlock, QFontMetricsF, QPainter, QLinearGradient,
                        QPen, QResizeEvent, QCursor, QTextCursor)
from qtpy.QtWidgets import QApplication

# Local imports
from spyder.plugins.editor.api.panel import Panel
from spyder.config.gui import is_dark_interface
from spyder.plugins.editor.api.decoration import TextDecoration, DRAW_ORDERS
from spyder.plugins.editor.panels.utils import FoldingRegion
from spyder.plugins.editor.utils.editor import (TextHelper, DelayJobRunner,
                                                drift_color)
from spyder.utils.icon_manager import ima
from spyder.widgets.mixins import EOL_SYMBOLS


class FoldingPanel(Panel):

    """
    Displays the document outline and lets the user collapse/expand blocks.

    The data represented by the panel come from the Language Server Protocol
    invoked via the CodeEditor. This panel stores information about both
    folding regions and their folding state.
    """
    #: signal emitted when a fold trigger state has changed, parameters are
    #: the concerned text block and the new state (collapsed or not).
    trigger_state_changed = Signal(QTextBlock, bool)
    collapse_all_triggered = Signal()
    expand_all_triggered = Signal()

    def __init__(self):
        Panel.__init__(self)

        self.collapsed_icon = ima.icon('folding.arrow_right')
        self.uncollapsed_icon = ima.icon('folding.arrow_down')

        self._block_nbr = -1
        self._highlight_caret = False
        self.highlight_caret_scope = False

        #: the list of deco used to highlight the current fold region (
        #: surrounding regions are darker)
        self._scope_decos = []

        #: the list of folded block decorations
        self._block_decos = {}

        self.setMouseTracking(True)
        self.scrollable = True
        self._mouse_over_line = None
        self._current_scope = None
        self._display_folding = False
        self._key_pressed = False
        self._highlight_runner = DelayJobRunner(delay=250)
        self.current_tree = IntervalTree()
        self.root = FoldingRegion(None, None)
        self.folding_regions = {}
        self.folding_status = {}
        self.folding_levels = {}
        self.folding_nesting = {}

    @property
    def highlight_caret_scope(self):
        """
        True to highlight the caret scope automatically.

        (Similar to the ``Highlight blocks in Qt Creator``.

        Default is False.
        """
        return self._highlight_caret

    @highlight_caret_scope.setter
    def highlight_caret_scope(self, value):
        if value != self._highlight_caret:
            self._highlight_caret = value
            if self.editor:
                if value:
                    self._block_nbr = -1
                    self.editor.cursorPositionChanged.connect(
                        self._highlight_caret_scope)
                else:
                    self._block_nbr = -1
                    self.editor.cursorPositionChanged.disconnect(
                        self._highlight_caret_scope)
                for clone in self.editor.clones:
                    try:
                        clone.modes.get(
                            self.__class__).highlight_caret_scope = value
                    except KeyError:
                        # this should never happen since we're working with
                        # clones
                        pass

    def update_folding(self, folding_info):
        """Update folding panel folding ranges."""
        if folding_info is None:
            return

        (self.current_tree, self.root,
         self.folding_regions, self.folding_nesting,
         self.folding_levels, self.folding_status) = folding_info

        self._clear_block_decos()
        self.update()

    def sizeHint(self):
        """Returns the widget size hint (based on the editor font size) """
        fm = QFontMetricsF(self.editor.font())
        size_hint = QSize(ceil(fm.height()), ceil(fm.height()))
        if size_hint.width() > 16:
            size_hint.setWidth(16)
        return size_hint

    def _draw_collapsed_indicator(self, line_number, top_position, block,
                                  painter, mouse_hover=False):
        if line_number in self.folding_regions:
            collapsed = self.folding_status[line_number]

            if not mouse_hover:
                self._draw_fold_indicator(top_position, collapsed, painter)

            if collapsed:
                if mouse_hover:
                    self._draw_fold_indicator(top_position, collapsed, painter)
            elif not mouse_hover:
                for deco_line in list(self._block_decos.keys()):
                    deco = self._block_decos[deco_line]

                    # Check if the block decoration has been removed, it
                    # might have been unfolded by the parent
                    # editor/document in the case of cloned editor
                    if deco.block == block:
                        # remove it and
                        self._block_decos.pop(deco_line)
                        self.editor.decorations.remove(deco, key='folded')
                        del deco
                        break

    def highlight_folded_regions(self):
        """Highlight folded regions on the editor's visible buffer."""
        first_block_nb, last_block_nb = self.editor.get_buffer_block_numbers()

        # This can happen at startup, and when it does, we don't need to move
        # pass this point.
        if first_block_nb == last_block_nb:
            return

        for block_number in range(first_block_nb, last_block_nb):
            block = self.editor.document().findBlockByNumber(block_number)
            line_number = block_number + 1

            if line_number in self.folding_regions:
                collapsed = self.folding_status[line_number]

                # Check if a block is folded by UI inspection.
                # This is necesary because the algorithm that detects the
                # currently folded regions may fail, for instance, after
                # pasting a big chunk of code.
                ui_collapsed = (
                    block.isVisible() and not block.next().isVisible()
                )

                if collapsed != ui_collapsed:
                    collapsed = ui_collapsed
                    self.folding_status[line_number] = ui_collapsed

                if collapsed:
                    # Check if the block already has a decoration,
                    # it might have been folded by the parent
                    # editor/document in the case of cloned editor
                    for deco_line in self._block_decos:
                        deco = self._block_decos[deco_line]
                        if deco.block == block:
                            # no need to add a deco, just go to the
                            # next block
                            break
                    else:
                        line_end = self.folding_regions[line_number]
                        self._add_fold_decoration(block, line_end)

    def paintEvent(self, event):
        """
        Paint fold indicators on the folding panel and possible folding region
        background on the editor.
        """
        super().paintEvent(event)
        painter = QPainter(self)
        self.paint_cell(painter)

        # Draw collapsed indicators
        if not self._display_folding and not self._key_pressed:
            for top_position, line_number, block in self.editor.visible_blocks:
                if self.folding_status.get(line_number):
                    self._draw_collapsed_indicator(
                        line_number, top_position, block,
                        painter, mouse_hover=True)
            return

        # Draw background over the selected non collapsed fold region
        if self._mouse_over_line is not None:
            block = self.editor.document().findBlockByNumber(
                self._mouse_over_line)
            try:
                self._draw_fold_region_background(block, painter)
            except (ValueError, KeyError):
                # Catching the KeyError above is necessary to avoid
                # issue spyder-ide/spyder#10918.
                # It happens when users have the mouse on top of the
                # folding panel and make some text modifications
                # that trigger a folding recomputation.
                pass

        # Draw all fold indicators
        for top_position, line_number, block in self.editor.visible_blocks:
            self._draw_collapsed_indicator(
                line_number, top_position, block, painter, mouse_hover=False)

    def _draw_fold_region_background(self, block, painter):
        """
        Draw the fold region when the mouse is over and non collapsed
        indicator.

        :param top: Top position
        :param block: Current block.
        :param painter: QPainter
        """
        th = TextHelper(self.editor)
        start = block.blockNumber()
        end = self.folding_regions[start]
        if start > 0:
            top = th.line_pos_from_number(start)
        else:
            top = 0
        bottom = th.line_pos_from_number(end)
        h = bottom - top
        if h == 0:
            h = self.sizeHint().height()
        w = self.sizeHint().width()
        self._draw_rect(QRectF(0, top, w, h), painter)

    def _draw_rect(self, rect, painter):
        """
        Draw the background rectangle using the current style primitive color.

        :param rect: The fold zone rect to draw

        :param painter: The widget's painter.
        """
        c = self.editor.sideareas_color
        grad = QLinearGradient(rect.topLeft(), rect.topRight())

        if is_dark_interface():
            grad.setColorAt(0, c.lighter(110))
            grad.setColorAt(1, c.lighter(130))
            outline = c.darker(100)
        else:
            grad.setColorAt(0, c.darker(105))
            grad.setColorAt(1, c.darker(115))
            outline = c.lighter(110)

        painter.fillRect(rect, grad)
        painter.setPen(QPen(outline))
        painter.drawLine(rect.topLeft() +
                         QPointF(1, 0),
                         rect.topRight() -
                         QPointF(1, 0))
        painter.drawLine(rect.bottomLeft() +
                         QPointF(1, 0),
                         rect.bottomRight() -
                         QPointF(1, 0))
        painter.drawLine(rect.topRight() +
                         QPointF(0, 1),
                         rect.bottomRight() -
                         QPointF(0, 1))
        painter.drawLine(rect.topLeft() +
                         QPointF(0, 1),
                         rect.bottomLeft() -
                         QPointF(0, 1))

    def _draw_fold_indicator(self, top, collapsed, painter):
        """
        Draw the fold indicator/trigger (arrow).

        :param top: Top position
        :param collapsed: Whether the trigger is collapsed or not.
        :param painter: QPainter
        """
        rect = QRect(
            0, top, self.sizeHint().width() + 2, self.sizeHint().height() + 2
        )

        if collapsed:
            icon = self.collapsed_icon
        else:
            icon = self.uncollapsed_icon

        icon.paint(painter, rect)

    def find_parent_scope(self, block):
        """Find parent scope, if the block is not a fold trigger."""
        block_line = block.blockNumber()
        if block_line not in self.folding_regions:
            for start_line in self.folding_regions:
                end_line = self.folding_regions[start_line]
                if end_line > block_line:
                    if start_line < block_line:
                        block = self.editor.document().findBlockByNumber(
                            start_line)
                        break
        return block

    def _clear_scope_decos(self):
        """Clear scope decorations (on the editor)"""
        for deco in self._scope_decos:
            self.editor.decorations.remove(deco)
        self._scope_decos[:] = []

    def _get_scope_highlight_color(self):
        """
        Gets the base scope highlight color (derivated from the editor
        background)

        For lighter themes will be a darker color,
        and for darker ones will be a lighter color
        """
        color = self.editor.sideareas_color

        if is_dark_interface():
            color = drift_color(color, 130)
        else:
            color = drift_color(color, 105)

        return color

    def _decorate_block(self, start, end):
        """
        Create a decoration and add it to the editor.

        Args:
            start (int) start line of the decoration
            end (int) end line of the decoration
        """
        color = self._get_scope_highlight_color()
        draw_order = DRAW_ORDERS.get('folding_areas')
        d = TextDecoration(self.editor.document(),
                           start_line=max(0, start - 1),
                           end_line=end,
                           draw_order=draw_order)
        d.set_background(color)
        d.set_full_width(True, clear=False)
        self.editor.decorations.add(d)
        self._scope_decos.append(d)

    def _highlight_block(self, block):
        """
        Highlights the current fold scope.

        :param block: Block that starts the current fold scope.
        """
        block_line = block.blockNumber()
        end_line = self.folding_regions[block_line]

        scope = (block_line, end_line)
        if (self._current_scope is None or self._current_scope != scope):
            self._current_scope = scope
            self._clear_scope_decos()
            # highlight current scope with darker or lighter color
            start, end = scope
            if not self.folding_status[start]:
                self._decorate_block(start, end)

    def mouseMoveEvent(self, event):
        """
        Detect mouser over indicator and highlight the current scope in the
        editor (up and down decoration arround the foldable text when the mouse
        is over an indicator).

        :param event: event
        """
        super(FoldingPanel, self).mouseMoveEvent(event)
        th = TextHelper(self.editor)
        line = th.line_nbr_from_position(event.pos().y())

        if line >= 0:
            block = self.editor.document().findBlockByNumber(line)
            block = self.find_parent_scope(block)
            line_number = block.blockNumber()

            if line_number in self.folding_regions:
                if self._mouse_over_line is None:
                    # mouse enter fold scope
                    QApplication.setOverrideCursor(
                        QCursor(Qt.PointingHandCursor))
                if (self._mouse_over_line != block.blockNumber() and
                        self._mouse_over_line is not None):
                    # fold scope changed, a previous block was highlighter so
                    # we quickly update our highlighting
                    self._mouse_over_line = block.blockNumber()
                    try:
                        self._highlight_block(block)
                    except KeyError:
                        # Catching the KeyError above is necessary to avoid
                        # issue spyder-ide/spyder#13230.
                        pass
                else:
                    # same fold scope, request highlight
                    self._mouse_over_line = block.blockNumber()
                    try:
                        self._highlight_runner.request_job(
                            self._highlight_block, block)
                    except KeyError:
                        # Catching the KeyError above is necessary to avoid
                        # issue spyder-ide/spyder#11291.
                        pass
                self._highight_block = block
            else:
                # no fold scope to highlight, cancel any pending requests
                self._highlight_runner.cancel_requests()
                self._mouse_over_line = None
                QApplication.restoreOverrideCursor()

            self.repaint()

    def enterEvent(self, event):
        self._display_folding = True
        self.repaint()

    def leaveEvent(self, event):
        """
        Removes scope decorations and background from the editor and the panel
        if highlight_caret_scope, else simply update the scope decorations to
        match the caret scope.
        """
        super(FoldingPanel, self).leaveEvent(event)
        QApplication.restoreOverrideCursor()
        self._highlight_runner.cancel_requests()
        if not self.highlight_caret_scope:
            self._clear_scope_decos()
            self._mouse_over_line = None
            self._current_scope = None
        else:
            self._block_nbr = -1
            self._highlight_caret_scope()
        self.editor.repaint()
        self._display_folding = False

    def _add_fold_decoration(self, block, end_line):
        """
        Add fold decorations (boxes arround a folded block in the editor
        widget).
        """
        start_line = block.blockNumber()
        text = self.editor.get_text_region(start_line + 1, end_line)
        draw_order = DRAW_ORDERS.get('folded_regions')

        deco = TextDecoration(block, draw_order=draw_order)
        deco.signals.clicked.connect(self._on_fold_deco_clicked)
        deco.tooltip = text
        deco.block = block
        deco.select_line()
        deco.set_background(self._get_scope_highlight_color())
        deco.set_full_width(flag=True, clear=True)
        self._block_decos[start_line] = deco
        self.editor.decorations.add(deco, key='folded')

    def _get_block_until_line(self, block, end_line):
        while block.blockNumber() <= end_line and block.isValid():
            block.setVisible(False)
            block = block.next()
        return block

    def fold_region(self, block, start_line, end_line):
        """Fold region spanned by `start_line` and `end_line`."""
        # Note: The block passed to this method is the first one that needs to
        # be hidden.
        initial_block = self.editor.document().findBlockByNumber(
            start_line - 1)
        self._add_fold_decoration(initial_block, end_line)

        while block.blockNumber() < end_line and block.isValid():
            block.setVisible(False)
            block = block.next()

    def unfold_region(self, block, start_line, end_line):
        """Unfold region spanned by `start_line` and `end_line`."""
        if start_line - 1 in self._block_decos:
            deco = self._block_decos[start_line - 1]
            self._block_decos.pop(start_line - 1)
            self.editor.decorations.remove(deco, key='folded')

        while block.blockNumber() < end_line and block.isValid():
            current_line = block.blockNumber()
            block.setVisible(True)
            get_next = True

            if (
                current_line in self.folding_regions
                and current_line != start_line
            ):
                block_end = self.folding_regions[current_line]
                if self.folding_status[current_line]:
                    # Skip setting visible blocks until the block is done
                    get_next = False
                    block = self._get_block_until_line(block, block_end - 1)

            if get_next:
                block = block.next()

    def toggle_fold_trigger(self, block):
        """
        Toggle a fold trigger block (expand or collapse it).

        :param block: The QTextBlock to expand/collapse
        """
        start_line = block.blockNumber()

        if start_line not in self.folding_regions:
            return

        end_line = self.folding_regions[start_line]
        if self.folding_status[start_line]:
            self.unfold_region(block, start_line, end_line)
            self.folding_status[start_line] = False
            if self._mouse_over_line is not None:
                self._decorate_block(start_line, end_line)
        else:
            self.fold_region(block, start_line, end_line)
            self.folding_status[start_line] = True
            self._clear_scope_decos()

        self._refresh_editor_and_scrollbars()

    def mousePressEvent(self, event):
        """Folds/unfolds the pressed indicator if any."""
        if self._mouse_over_line is not None:
            block = self.editor.document().findBlockByNumber(
                self._mouse_over_line)
            self.toggle_fold_trigger(block)

    def _on_fold_deco_clicked(self, deco):
        """Unfold a folded block that has just been clicked by the user"""
        self.toggle_fold_trigger(deco.block)

    def on_state_changed(self, state):
        """
        On state changed we (dis)connect to the cursorPositionChanged signal
        """
        if state:
            self.editor.sig_key_pressed.connect(self._on_key_pressed)
            self.editor.sig_delete_requested.connect(self._expand_selection)

            if self._highlight_caret:
                self.editor.cursorPositionChanged.connect(
                    self._highlight_caret_scope)
                self._block_nbr = -1

            self.editor.new_text_set.connect(self._clear_block_decos)
        else:

            self.editor.sig_key_pressed.disconnect(self._on_key_pressed)
            self.editor.sig_delete_requested.disconnect(self._expand_selection)

            if self._highlight_caret:
                self.editor.cursorPositionChanged.disconnect(
                    self._highlight_caret_scope)
                self._block_nbr = -1

            self.editor.new_text_set.disconnect(self._clear_block_decos)

    def _in_folded_block(self):
        """Check if the current block is folded."""
        cursor = self.editor.textCursor()

        if cursor.hasSelection():
            block_start = self.editor.document().findBlock(
                cursor.selectionStart()
            )
            block_end = self.editor.document().findBlock(cursor.selectionEnd())

            if (
                # The start block needs to be among the folded ones.
                block_start.blockNumber() in self._block_decos
                # This covers the case when there's some text selected in the
                # folded line or when it's selected in its entirety. For the
                # latter, Qt returns the next block as the final one, which
                # is not visible.
                and (block_start == block_end or not block_end.isVisible())
            ):
                return True
            else:
                return False
        else:
            current_block = self.editor.document().findBlock(cursor.position())
            return current_block.blockNumber() in self._block_decos

    def _on_key_pressed(self, event):
        """
        Handle key press events in order to select a whole folded scope if the
        user wants to remove it.

        Notes
        -----
        We don't handle Key_Delete here because it's behind a shortcut in
        CodeEditor. So, the event associated to that key doesn't go through its
        keyPressEvent.

        Instead, CodeEditor emits sig_delete_requested in the method that gets
        called when Key_Delete is pressed, and in several other places, which
        is handled by _expand_selection below.
        """
        # This doesn't apply if there are not folded regions
        if not self._block_decos:
            return

        if self._in_folded_block():
            # We prevent the following events to change folded blocks to make
            # them appear as read-only to users.
            # See the last comments in spyder-ide/spyder#21669 for the details
            # of this decision.
            if (
                # When Tab or Shift+Tab are pressed
                event.key() in [Qt.Key_Tab, Qt.Key_Backtab]
                # When text is trying to be written
                or event.text() and event.key() != Qt.Key_Backspace
            ):
                event.accept()
                return

        delete_pressed = event.key() == Qt.Key_Backspace

        enter_pressed = False
        cursor = self.editor.textCursor()
        if cursor.hasSelection():
            if event.key() == Qt.Key_Return:
                enter_pressed = True

        # Delete a folded scope when pressing delete or enter
        if delete_pressed or enter_pressed:
            self._expand_selection()

    def _expand_selection(self):
        """
        Expand selection to encompass a whole folded scope in case the
        current selection starts and/or ends in one, or the cursor is over a
        block deco.
        """
        if not self._block_decos:
            return

        cursor = self.editor.textCursor()
        self._key_pressed = True

        # If there's no selected text, select the current line but only if
        # it corresponds to a block deco. That allows us to remove the folded
        # region associated to it when typing Delete or Backspace on the line.
        # Otherwise, the editor ends up in an inconsistent state.
        if not cursor.hasSelection():
            current_block = self.editor.document().findBlock(cursor.position())

            if current_block.blockNumber() in self._block_decos:
                cursor.select(QTextCursor.LineUnderCursor)
            else:
                self._key_pressed = False
                return

        # Get positions to check if we need to expand the current selection to
        # cover a folded region too.
        start_pos = cursor.selectionStart()
        end_pos = cursor.selectionEnd()

        # A selection can end in an eol when calling CodeEditor.delete_line,
        # for instance. In that case, we need to remove it for the code below
        # to work as expected.
        if cursor.selectedText()[-1] in EOL_SYMBOLS:
            end_pos -= 1

        positions_to_check = (start_pos, end_pos)
        for pos in positions_to_check:
            block = self.editor.document().findBlock(pos)
            start_line = block.blockNumber() + 1

            if (
                start_line in self.folding_regions
                and self.folding_status[start_line]
            ):
                end_line = self.folding_regions[start_line]

                if cursor.hasSelection():
                    tc = TextHelper(self.editor).select_lines(
                        start_line, end_line)
                    tc.movePosition(tc.MoveOperation.NextBlock,
                                    tc.MoveMode.KeepAnchor)

                    if tc.selectionStart() > cursor.selectionStart():
                        start = cursor.selectionStart()
                    else:
                        start = tc.selectionStart()

                    if tc.selectionEnd() < cursor.selectionEnd():
                        end = cursor.selectionEnd()
                    else:
                        end = tc.selectionEnd()

                    tc.setPosition(start)
                    tc.setPosition(end, tc.KeepAnchor)

                    self.editor.setTextCursor(tc)

        self._update_block_decos(start_pos, end_pos)
        self._key_pressed = False

    def _refresh_editor_and_scrollbars(self):
        """
        Refrehes editor content and scollbars.

        We generate a fake resize event to refresh scroll bar.

        We have the same problem as described here:
        http://www.qtcentre.org/threads/44803 and we apply the same solution
        (don't worry, there is no visual effect, the editor does not grow up
        at all, even with a value = 500)
        """
        TextHelper(self.editor).mark_whole_doc_dirty()
        self.editor.repaint()
        s = self.editor.size()
        s.setWidth(s.width() + 1)
        self.editor.resizeEvent(QResizeEvent(self.editor.size(), s))

    def collapse_all(self):
        """
        Collapses all triggers and makes all blocks with fold level > 0
        invisible.
        """
        self._clear_block_decos()
        block = self.editor.document().firstBlock()
        while block.isValid():
            line_number = block.blockNumber()
            if line_number in self.folding_regions:
                end_line = self.folding_regions[line_number]
                self.fold_region(block, line_number, end_line)
            block = block.next()
        self._refresh_editor_and_scrollbars()
        tc = self.editor.textCursor()
        tc.movePosition(tc.Start)
        self.editor.setTextCursor(tc)
        self.collapse_all_triggered.emit()

    def _clear_block_decos(self):
        """Clear the folded block decorations."""
        self.editor.decorations.remove_key('folded')
        self._block_decos = {}

    def _update_block_decos(self, start_pos, end_pos):
        """
        Update block decorations in case some are going to be removed by the
        user.

        Parameters
        ----------
        start_pos: int
            Start cursor position of the selection that's going to remove or
            replace text in the editor
        end_pos: int
            End cursor position of the same selection.
        """
        start_line = self.editor.document().findBlock(start_pos).blockNumber()
        end_line = self.editor.document().findBlock(end_pos).blockNumber()

        for deco_line in self._block_decos.copy():
            if start_line <= deco_line <= end_line:
                deco = self._block_decos[deco_line]
                self._block_decos.pop(deco_line)
                self.editor.decorations.remove(deco, key='folded')
                self.folding_status[deco_line + 1] = False

    def expand_all(self):
        """Expands all fold triggers."""
        block = self.editor.document().firstBlock()
        while block.isValid():
            line_number = block.BlockNumber()
            if line_number in self.folding_regions:
                end_line = self.folding_regions[line_number]
                self.unfold_region(block, line_number, end_line)
            block = block.next()
        self._clear_block_decos()
        self._refresh_editor_and_scrollbars()
        self.expand_all_triggered.emit()

    def _highlight_caret_scope(self):
        """
        Highlight the scope of the current caret position.

        This get called only if :attr:`
        spyder.widgets.panels.FoldingPanel.highlight_care_scope` is True.
        """
        cursor = self.editor.textCursor()
        block_nbr = cursor.blockNumber()
        if self._block_nbr != block_nbr:
            block = self.find_parent_scope(cursor.block())
            line_number = block.blockNumber()
            if line_number in self.folding_regions:
                self._mouse_over_line = block.blockNumber()
                try:
                    self._highlight_block(block)
                except KeyError:
                    # Catching the KeyError above is necessary to avoid
                    # issue spyder-ide/spyder#13230.
                    pass
            else:
                self._clear_scope_decos()
        self._block_nbr = block_nbr



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Helper widgets.
"""

# Standard imports
from __future__ import annotations
import re
import textwrap
from typing import Callable

# Third party imports
import qtawesome as qta
import qstylizer.style
from qtpy import PYQT5
from qtpy.QtCore import (
    QEvent,
    QPoint,
    QRegularExpression,
    QSize,
    QSortFilterProxyModel,
    Qt,
    QTimer,
    Signal,
)
from qtpy.QtGui import (
    QAbstractTextDocumentLayout,
    QColor,
    QFontMetrics,
    QIcon,
    QPainter,
    QRegularExpressionValidator,
    QTextDocument,
)
from qtpy.QtWidgets import (
    QAction,
    QApplication,
    QCheckBox,
    QLineEdit,
    QMessageBox,
    QSpacerItem,
    QStyle,
    QStyledItemDelegate,
    QStyleOptionFrame,
    QStyleOptionViewItem,
    QTableView,
    QToolButton,
    QToolTip,
    QVBoxLayout,
    QWidget,
    QHBoxLayout,
    QLabel,
    QFrame,
    QItemDelegate,
)

# Local imports
from spyder.api.fonts import SpyderFontType, SpyderFontsMixin
from spyder.api.widgets.mixins import SvgToScaledPixmap
from spyder.api.widgets.comboboxes import SpyderComboBox
from spyder.config.base import _
from spyder.utils.icon_manager import ima
from spyder.utils.stringmatching import get_search_regex
from spyder.utils.palette import SpyderPalette
from spyder.utils.stylesheet import AppStyle


# Valid finder chars. To be improved
VALID_ACCENT_CHARS = "ÁÉÍOÚáéíúóàèìòùÀÈÌÒÙâêîôûÂÊÎÔÛäëïöüÄËÏÖÜñÑ"
VALID_FINDER_CHARS = r"[A-Za-z\s{0}]".format(VALID_ACCENT_CHARS)


class MessageCheckBox(QMessageBox):
    """
    A QMessageBox derived widget that includes a QCheckBox aligned to the right
    under the message and on top of the buttons.
    """
    def __init__(self, *args, **kwargs):
        super(MessageCheckBox, self).__init__(*args, **kwargs)

        self.setWindowModality(Qt.NonModal)
        self._checkbox = QCheckBox(self)

        # Set layout to include checkbox
        size = 9
        check_layout = QVBoxLayout()
        check_layout.addItem(QSpacerItem(size, size))
        check_layout.addWidget(self._checkbox, 0, Qt.AlignRight)
        check_layout.addItem(QSpacerItem(size, size))

        # Access the Layout of the MessageBox to add the Checkbox
        layout = self.layout()
        if PYQT5:
            layout.addLayout(check_layout, 1, 2)
        else:
            layout.addLayout(check_layout, 1, 1)

    # --- Public API
    # Methods to access the checkbox
    def is_checked(self):
        return self._checkbox.isChecked()

    def set_checked(self, value):
        return self._checkbox.setChecked(value)

    def set_check_visible(self, value):
        self._checkbox.setVisible(value)

    def is_check_visible(self):
        self._checkbox.isVisible()

    def checkbox_text(self):
        self._checkbox.text()

    def set_checkbox_text(self, text):
        self._checkbox.setText(text)


class HTMLDelegate(QStyledItemDelegate):
    """
    With this delegate, a QListWidgetItem or a QTableItem can render HTML.

    Taken from https://stackoverflow.com/a/5443112/2399799
    """

    def __init__(self, parent, margin=0, wrap_text=False, align_vcenter=False):
        super(HTMLDelegate, self).__init__(parent)
        self._margin = margin
        self._wrap_text = wrap_text
        self._hovered_row = -1
        self._align_vcenter = align_vcenter

    def _prepare_text_document(self, option, index):
        # This logic must be shared between paint and sizeHint for consistency
        options = QStyleOptionViewItem(option)
        self.initStyleOption(options, index)

        doc = QTextDocument()
        doc.setDocumentMargin(self._margin)
        doc.setHtml(options.text)
        if self._wrap_text:
            # The -25 here is used to avoid the text to go totally up to the
            # right border of the widget that contains the delegate, which
            # doesn't look good.
            doc.setTextWidth(option.rect.width() - 25)

        return options, doc

    def on_hover_index_changed(self, index):
        """
        This can be used by a widget that inherits from HoverRowsTableView to
        connect its sig_hover_index_changed signal to this method to paint an
        entire row when it's hovered.
        """
        self._hovered_row = index.row()

    def paint(self, painter, option, index):
        options, doc = self._prepare_text_document(option, index)

        style = (QApplication.style() if options.widget is None
                 else options.widget.style())
        options.text = ""

        # This paints the entire row associated to the delegate when it's
        # hovered and the table that holds it informs it what's the current
        # row (see HoverRowsTableView for an example).
        if index.row() == self._hovered_row:
            painter.fillRect(
                options.rect, QColor(SpyderPalette.COLOR_BACKGROUND_3)
            )

        # Note: We need to pass the options widget as an argument of
        # drawCrontol to make sure the delegate is painted with a style
        # consistent with the widget in which it is used.
        # See spyder-ide/spyder#10677.
        style.drawControl(QStyle.CE_ItemViewItem, options, painter,
                          options.widget)

        ctx = QAbstractTextDocumentLayout.PaintContext()

        textRect = style.subElementRect(QStyle.SE_ItemViewItemText,
                                        options, None)
        painter.save()

        # Adjustments for the file switcher
        if hasattr(options.widget, 'files_list'):
            if options.widget.files_list:
                painter.translate(textRect.topLeft() + QPoint(4, 4))
            else:
                painter.translate(textRect.topLeft() + QPoint(2, 4))
        else:
            if not self._align_vcenter:
                painter.translate(textRect.topLeft() + QPoint(0, -3))

        # Center text vertically if requested.
        # Take from https://stackoverflow.com/a/32911270/438386
        if self._align_vcenter:
            doc.setTextWidth(option.rect.width())
            offset_y = (option.rect.height() - doc.size().height()) / 2
            painter.translate(options.rect.x(), options.rect.y() + offset_y)
            doc.drawContents(painter)

        # Type check: Prevent error in PySide where using
        # doc.documentLayout().draw() may fail because doc.documentLayout()
        # returns an object of type QtGui.QStandardItem (for whatever reason).
        docLayout = doc.documentLayout()
        if type(docLayout) is QAbstractTextDocumentLayout:
            docLayout.draw(painter, ctx)

        painter.restore()

    def sizeHint(self, option, index):
        __, doc = self._prepare_text_document(option, index)
        return QSize(round(doc.idealWidth()), round(doc.size().height() - 2))


class ItemDelegate(QStyledItemDelegate):
    def __init__(self, parent):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        options = QStyleOptionViewItem(option)
        self.initStyleOption(options, index)

        style = (QApplication.style() if options.widget is None
                 else options.widget.style())

        doc = QTextDocument()
        doc.setDocumentMargin(0)
        doc.setHtml(options.text)

        options.text = ""
        style.drawControl(QStyle.CE_ItemViewItem, options, painter)

        ctx = QAbstractTextDocumentLayout.PaintContext()

        textRect = style.subElementRect(QStyle.SE_ItemViewItemText,
                                        options, None)
        painter.save()

        painter.translate(textRect.topLeft())
        painter.setClipRect(textRect.translated(-textRect.topLeft()))
        doc.documentLayout().draw(painter, ctx)
        painter.restore()

    def sizeHint(self, option, index):
        options = QStyleOptionViewItem(option)
        self.initStyleOption(options, index)

        doc = QTextDocument()
        doc.setHtml(options.text)
        doc.setTextWidth(options.rect.width())

        return QSize(int(doc.idealWidth()), int(doc.size().height()))


class IconLineEdit(QLineEdit):
    """
    Custom QLineEdit that includes an icon representing a validation for its
    text and can also elide it.
    """

    def __init__(self, parent, elide_text=False, ellipsis_place=Qt.ElideLeft):
        super().__init__(parent)

        self.elide_text = elide_text
        self.ellipsis_place = ellipsis_place
        self._status = True
        self._status_set = True
        self._focus_in = False
        self._valid_icon = ima.icon('todo')
        self._invalid_icon = ima.icon('warning')
        self._set_icon = ima.icon('todo_list')
        self._refresh()
        self._paint_count = 0
        self._icon_visible = False

    def _refresh(self):
        """
        This makes space for the right validation icons after focus is given to
        the widget.
        """
        padding = self.height()
        if self.elide_text and not self._focus_in:
            padding = 0

        css = qstylizer.style.StyleSheet()
        css.QLineEdit.setValues(
            border='none',
            paddingLeft=f"{AppStyle.MarginSize}px",
            paddingRight=f"{padding}px",
            # This is necessary to correctly center the text
            paddingTop="0px",
            paddingBottom="0px",
            # Prevent jitter when giving focus to the line edit
            marginLeft=f"-{2 if self._focus_in else 0}px",
        )

        self.setStyleSheet(css.toString())
        self.update()

    def hide_status_icon(self):
        """Show the status icon."""
        self._icon_visible = False
        self.repaint()
        self.update()

    def show_status_icon(self):
        """Hide the status icon."""
        self._icon_visible = True
        self.repaint()
        self.update()

    def update_status(self, value, value_set):
        """Update the status and set_status to update the icons to display."""
        self._status = value
        self._status_set = value_set
        self.repaint()
        self.update()

    def paintEvent(self, event):
        """
        Include a validation icon to the left of the line edit and elide text
        if requested.
        """
        # Elide text if requested.
        # See PR spyder-ide/spyder#20005 for context.
        if self.elide_text and not self._focus_in:
            # This code is taken for the most part from the
            # AmountEdit.paintEvent method, part of the Electrum project. See
            # the Electrum entry in our NOTICE.txt file for the details.
            # Licensed under the MIT license.
            painter = QPainter(self)
            option = QStyleOptionFrame()
            self.initStyleOption(option)
            text_rect = self.style().subElementRect(
                QStyle.SE_LineEditContents, option, self)
            text_rect.adjust(0, 0, -2, 0)
            fm = QFontMetrics(self.font())
            text = fm.elidedText(self.text(), self.ellipsis_place,
                                 text_rect.width())
            painter.setPen(QColor(SpyderPalette.COLOR_TEXT_1))
            painter.drawText(text_rect, int(Qt.AlignLeft | Qt.AlignVCenter),
                             text)
            return

        super().paintEvent(event)
        painter = QPainter(self)
        rect = self.geometry()
        space = int((rect.height())/6)
        h = rect.height() - space
        w = rect.width() - h

        if self._icon_visible:
            if self._status and self._status_set:
                pixmap = self._set_icon.pixmap(h, h)
            elif self._status:
                pixmap = self._valid_icon.pixmap(h, h)
            else:
                pixmap = self._invalid_icon.pixmap(h, h)

            painter.drawPixmap(w, 1, pixmap)

        # Small hack to guarantee correct padding on Spyder start
        if self._paint_count < 5:
            self._paint_count += 1
            self._refresh()

    def focusInEvent(self, event):
        """Reimplemented to know when this widget has received focus."""
        self._focus_in = True
        self._refresh()
        super().focusInEvent(event)

    def focusOutEvent(self, event):
        """Reimplemented to know when this widget has lost focus."""
        self._focus_in = False
        self._refresh()
        super().focusOutEvent(event)


class ClearLineEdit(QLineEdit):
    """QLineEdit with a clear button."""

    def __init__(self, parent, reposition_button=False):
        super().__init__(parent)

        # Add button to clear text inside the line edit.
        self.clear_action = QAction(self)
        self.clear_action.setIcon(ima.icon('clear_text'))
        self.clear_action.setToolTip(_('Clear text'))
        self.clear_action.triggered.connect(self.clear)
        self.addAction(self.clear_action, QLineEdit.TrailingPosition)

        # Button that corresponds to the clear_action above
        self.clear_button = self.findChildren(QToolButton)[0]

        # Hide clear_action by default because lineEdit is empty when the
        # combobox is created, so it doesn't make sense to show it.
        self.clear_action.setVisible(False)

        # Signals
        self.textChanged.connect(self._on_text_changed)

        # Event filter
        if reposition_button:
            self.installEventFilter(self)

    def _on_text_changed(self, text):
        """Actions to take when text has changed on the line edit widget."""
        if text:
            self.clear_action.setVisible(True)
        else:
            self.clear_action.setVisible(False)

    def eventFilter(self, widget, event):
        """
        Event filter for this widget used to reduce the space between
        clear_button and the right border of the line edit.
        """
        if event.type() == QEvent.Paint:
            self.clear_button.move(self.width() - 22, self.clear_button.y())

        return super().eventFilter(widget, event)


class FinderLineEdit(ClearLineEdit):

    sig_hide_requested = Signal()
    sig_find_requested = Signal()

    def __init__(self, parent, regex_base=None, key_filter_dict=None):
        super().__init__(parent)
        self.key_filter_dict = key_filter_dict

        self._combobox = SpyderComboBox(self)
        self._is_shown = False

        if regex_base is not None:
            # Widget setup
            regex = QRegularExpression(regex_base + "{100}")
            self.setValidator(QRegularExpressionValidator(regex))

    def keyPressEvent(self, event):
        """Qt and FilterLineEdit Override."""
        key = event.key()
        if self.key_filter_dict is not None and key in self.key_filter_dict:
            self.key_filter_dict[key]()
        elif key in [Qt.Key_Escape]:
            self.sig_hide_requested.emit()
        elif key in [Qt.Key_Enter, Qt.Key_Return]:
            self.sig_find_requested.emit()
        else:
            super().keyPressEvent(event)

    def showEvent(self, event):
        """Adjustments when the widget is shown."""
        if not self._is_shown:
            height = self._combobox.size().height()
            self._combobox.hide()

            # Only set a min width so it grows with the parent's width.
            self.setMinimumWidth(AppStyle.FindMinWidth)

            # Set a fixed height so that it looks the same as our comboboxes.
            self.setMinimumHeight(height)
            self.setMaximumHeight(height)

            self._is_shown = True

        super().showEvent(event)


class FinderWidget(QWidget):

    sig_find_text = Signal(str)
    sig_hide_finder_requested = Signal()

    def __init__(self, parent, regex_base=None, key_filter_dict=None,
                 find_on_change=False):
        super().__init__(parent)

        # Parent is assumed to be a spyder widget
        self.text_finder = FinderLineEdit(
            self,
            regex_base=regex_base,
            key_filter_dict=key_filter_dict
        )
        self.text_finder.sig_find_requested.connect(self.do_find)
        if find_on_change:
            self.text_finder.textChanged.connect(self.do_find)
        self.text_finder.sig_hide_requested.connect(
            self.sig_hide_finder_requested)

        self.finder_close_button = QToolButton(self)
        self.finder_close_button.setIcon(ima.icon('DialogCloseButton'))
        self.finder_close_button.clicked.connect(
            self.sig_hide_finder_requested)

        finder_layout = QHBoxLayout()
        finder_layout.addWidget(self.finder_close_button)
        finder_layout.addWidget(self.text_finder)
        finder_layout.addStretch()
        finder_layout.setContentsMargins(
            2 * AppStyle.MarginSize,
            AppStyle.MarginSize,
            2 * AppStyle.MarginSize,
            0
        )
        self.setLayout(finder_layout)
        self.setVisible(False)

    def do_find(self):
        """Send text."""
        text = self.text_finder.text()
        if not text:
            text = ''
        self.sig_find_text.emit(text)

    def set_visible(self, visible):
        """Set visibility of widget."""
        self.setVisible(visible)
        if visible:
            self.text_finder.setFocus()
            self.do_find()
        else:
            self.sig_find_text.emit("")


class CustomSortFilterProxy(QSortFilterProxyModel):
    """Custom column filter based on regex."""

    def __init__(self, parent=None):
        super(CustomSortFilterProxy, self).__init__(parent)
        self._parent = parent
        self.pattern = re.compile(r'')

    def set_filter(self, text):
        """Set regular expression for filter."""
        self.pattern = get_search_regex(text)
        if self.pattern and text:
            self._parent.setSortingEnabled(False)
        else:
            self._parent.setSortingEnabled(True)
        self.invalidateFilter()

    def filterAcceptsRow(self, row_num, parent):
        """Qt override.

        Reimplemented from base class to allow the use of custom filtering.
        """
        model = self.sourceModel()
        name = model.row(row_num).name
        r = re.search(self.pattern, name)

        if r is None:
            return False
        else:
            return True


class PaneEmptyWidget(QFrame, SvgToScaledPixmap, SpyderFontsMixin):
    """Widget to show a pane/plugin functionality description."""

    def __init__(
        self,
        parent,
        icon_filename,
        text=None,
        description=None,
        top_stretch: int = 1,
        middle_stretch: int = 1,
        bottom_stretch: int = 0,
        spinner: bool = False,
    ):
        super().__init__(parent)

        # Attributes
        self._is_shown = False
        self._spin = None

        interface_font_size = self.get_font(
            SpyderFontType.Interface).pointSize()

        # Image (icon)
        image_label = QLabel(self)
        image_label.setPixmap(
            self.svg_to_scaled_pixmap(icon_filename, rescale=0.8)
        )
        image_label.setAlignment(Qt.AlignCenter)
        image_label_qss = qstylizer.style.StyleSheet()
        image_label_qss.QLabel.setValues(border="0px")
        image_label.setStyleSheet(image_label_qss.toString())

        # Main text
        if text is not None:
            text_label = QLabel(text, parent=self)
            text_label.setAlignment(Qt.AlignCenter)
            text_label.setWordWrap(True)
            text_label_qss = qstylizer.style.StyleSheet()
            text_label_qss.QLabel.setValues(
                fontSize=f"{interface_font_size + 5}pt", border="0px"
            )
            text_label.setStyleSheet(text_label_qss.toString())

        # Description text
        if description is not None:
            description_label = QLabel(description, parent=self)
            description_label.setAlignment(Qt.AlignCenter)
            description_label.setWordWrap(True)
            description_label_qss = qstylizer.style.StyleSheet()
            description_label_qss.QLabel.setValues(
                fontSize=f"{interface_font_size}pt",
                backgroundColor=SpyderPalette.COLOR_OCCURRENCE_3,
                border="0px",
                padding="20px",
            )
            description_label.setStyleSheet(description_label_qss.toString())

        # Setup layout
        pane_empty_layout = QVBoxLayout()

        # Add the top stretch
        pane_empty_layout.addStretch(top_stretch)

        # add the image_lebel (icon)
        pane_empty_layout.addWidget(image_label)

        # Display spinner if requested
        if spinner:
            spin_widget = qta.IconWidget()
            self._spin = qta.Spin(spin_widget, interval=3, autostart=False)
            spin_icon = qta.icon(
                "mdi.loading",
                color=ima.MAIN_FG_COLOR,
                animation=self._spin
            )

            spin_widget.setIconSize(QSize(32, 32))
            spin_widget.setIcon(spin_icon)
            spin_widget.setStyleSheet(image_label_qss.toString())
            spin_widget.setAlignment(Qt.AlignCenter)

            pane_empty_layout.addWidget(spin_widget)
            pane_empty_layout.addItem(QSpacerItem(20, 20))

        # If text, display text and stretch
        if text is not None:
            pane_empty_layout.addWidget(text_label)
            pane_empty_layout.addStretch(middle_stretch)

        # If description, display description
        if description is not None:
            pane_empty_layout.addWidget(description_label)

        pane_empty_layout.addStretch(bottom_stretch)
        pane_empty_layout.setContentsMargins(20, 0, 20, 20)
        self.setLayout(pane_empty_layout)

        # Setup border style
        self.setFocusPolicy(Qt.StrongFocus)
        self._apply_stylesheet(False)

    # ---- Public methods
    # -------------------------------------------------------------------------
    def setup(self, *args, **kwargs):
        """
        This method is needed when using this widget to show a "no connected
        console" message in plugins that inherit from ShellConnectMainWidget.
        """
        pass

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def showEvent(self, event):
        """Adjustments when the widget is shown."""
        if not self._is_shown:
            self._start_spinner()
            self._is_shown = True

        super().showEvent(event)

    def hideEvent(self, event):
        """Adjustments when the widget is hidden."""
        self._stop_spinner()
        self._is_shown = False
        super().hideEvent(event)

    def focusInEvent(self, event):
        self._apply_stylesheet(True)
        super().focusOutEvent(event)

    def focusOutEvent(self, event):
        self._apply_stylesheet(False)
        super().focusOutEvent(event)

    # ---- Private methods
    # -------------------------------------------------------------------------
    def _apply_stylesheet(self, focus):
        if focus:
            border_color = SpyderPalette.COLOR_ACCENT_3
        else:
            border_color = SpyderPalette.COLOR_BACKGROUND_4

        qss = qstylizer.style.StyleSheet()
        qss.QFrame.setValues(
            border=f'1px solid {border_color}',
            margin='0px',
            padding='0px',
            borderRadius=SpyderPalette.SIZE_BORDER_RADIUS
        )

        self.setStyleSheet(qss.toString())

    def _start_spinner(self):
        """
        Start spinner when requested, in case the widget has one (it's stopped
        by default).
        """
        if self._spin is not None:
            self._spin.start()

    def _stop_spinner(self):
        """Stop spinner when requested, in case the widget has one."""
        if self._spin is not None:
            self._spin.stop()


class HoverRowsTableView(QTableView):
    """QTableView subclass that can highlight an entire row when hovered."""

    sig_hover_index_changed = Signal(object)
    """
    This is emitted when the index that is currently hovered has changed.

    Parameters
    ----------
    index: object
        QModelIndex that has changed on hover.
    """

    def __init__(self, parent, custom_delegate=False):
        QTableView.__init__(self, parent)

        # For mouseMoveEvent
        self.setMouseTracking(True)

        # To remove background color for the hovered row when the mouse is not
        # over the widget.
        css = qstylizer.style.StyleSheet()
        css["QTableView::item"].setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_1
        )
        self._stylesheet = css.toString()

        if not custom_delegate:
            self._set_delegate()

    # ---- Qt methods
    def mouseMoveEvent(self, event):
        self._inform_hover_index_changed(event)

    def wheelEvent(self, event):
        super().wheelEvent(event)
        self._inform_hover_index_changed(event)

    def leaveEvent(self, event):
        super().leaveEvent(event)
        self.setStyleSheet(self._stylesheet)

    def enterEvent(self, event):
        super().enterEvent(event)
        self.setStyleSheet("")

    # ---- Private methods
    def _inform_hover_index_changed(self, event):
        index = self.indexAt(event.pos())
        if index.isValid():
            self.sig_hover_index_changed.emit(index)
            self.viewport().update()

    def _set_delegate(self):
        """
        Set a custom item delegate that can highlight the current row when
        hovered.
        """

        class HoverRowDelegate(QItemDelegate):

            def __init__(self, parent):
                super().__init__(parent)
                self._hovered_row = -1

            def on_hover_index_changed(self, index):
                self._hovered_row = index.row()

            def paint(self, painter, option, index):
                # This paints the entire row associated to the delegate when
                # it's hovered.
                if index.row() == self._hovered_row:
                    painter.fillRect(
                        option.rect, QColor(SpyderPalette.COLOR_BACKGROUND_3)
                    )

                super().paint(painter, option, index)

        self.setItemDelegate(HoverRowDelegate(self))
        self.sig_hover_index_changed.connect(
            self.itemDelegate().on_hover_index_changed
        )


class TipWidget(QLabel):
    """Icon widget to show information as a tooltip when clicked."""

    def __init__(
        self,
        tip_text: str,
        icon: QIcon,
        hover_icon: QIcon,
        size: int | None = None,
        wrap_text: bool = False
    ):
        super().__init__()

        if wrap_text:
            tip_text = '\n'.join(textwrap.wrap(tip_text, 50))
        self.tip_text = tip_text

        size = size if size is not None else AppStyle.ConfigPageIconSize
        self.icon = icon.pixmap(QSize(size, size))
        self.hover_icon = hover_icon.pixmap(QSize(size, size))

        # Timer to show the tip if users don't click on the widget
        self.tip_timer = QTimer(self)
        self.tip_timer.setInterval(300)
        self.tip_timer.setSingleShot(True)
        self.tip_timer.timeout.connect(self.show_tip)

        self.setPixmap(self.icon)
        self.setFixedWidth(size + 3)
        self.setFixedHeight(size + 3)

    def show_tip(self):
        """Show tooltip"""
        if not QToolTip.isVisible():
            QToolTip.showText(
                self.mapToGlobal(QPoint(self.width(), 15)),
                self.tip_text,
                self
            )

    def enterEvent(self, event):
        """
        Change cursor shape and set hover icon when the mouse is on the widget.
        """
        self.setCursor(Qt.PointingHandCursor)
        self.setPixmap(self.hover_icon)
        self.tip_timer.start()
        super().enterEvent(event)

    def leaveEvent(self, event):
        """Hide tooltip and restore icon when the mouse leaves the widget."""
        QToolTip.hideText()
        self.setPixmap(self.icon)
        self.tip_timer.stop()
        super().leaveEvent(event)

    def mouseReleaseEvent(self, event):
        """Show tooltip when the widget is clicked."""
        self.show_tip()


class ValidationLineEdit(QLineEdit):
    """Lineedit that validates its contents in focus out."""

    sig_focus_out = Signal(str)

    def __init__(
        self,
        validate_callback: Callable,
        validate_reason: str,
        text: str = "",
        parent: QWidget | None = None,
    ):
        super().__init__(text, parent)

        # Attributes
        self._validate_callback = validate_callback
        self._validate_reason = validate_reason
        self._is_show = False

        # Action to signal that text is not valid
        self.error_action = QAction(self)
        self.error_action.setIcon(ima.icon("error"))
        self.addAction(self.error_action, QLineEdit.TrailingPosition)

        # Signals
        self.sig_focus_out.connect(self._validate)

    def focusOutEvent(self, event):
        if self.text():
            self.sig_focus_out.emit(self.text())
        else:
            self.error_action.setVisible(False)
        super().focusOutEvent(event)

    def focusInEvent(self, event):
        self.error_action.setVisible(False)
        super().focusInEvent(event)

    def showEvent(self, event):
        if not self._is_show:
            self.error_action.setVisible(False)
            self._is_show = True
        super().showEvent(event)

    def _validate(self, text):
        if not self._validate_callback(text):
            self.error_action.setVisible(True)
            self.error_action.setToolTip(self._validate_reason)


def test_msgcheckbox():
    from spyder.utils.qthelpers import qapplication
    app = qapplication()  # noqa
    box = MessageCheckBox()
    box.setWindowTitle(_("Spyder updates"))
    box.setText("Testing checkbox")
    box.set_checkbox_text("Check for updates on startup?")
    box.setStandardButtons(QMessageBox.Ok)
    box.setDefaultButton(QMessageBox.Ok)
    box.setIcon(QMessageBox.Information)
    box.exec_()


if __name__ == '__main__':
    test_msgcheckbox()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
IPython Console mixins.
"""

# Local imports
from spyder.plugins.ipythonconsole.utils.kernel_handler import KernelHandler


class CachedKernelMixin:
    """Cached kernel mixin."""

    def __init__(self):
        super().__init__()
        self._cached_kernel_properties = None

    def close_cached_kernel(self):
        """Close the cached kernel."""
        if self._cached_kernel_properties is None:
            return
        kernel = self._cached_kernel_properties[-1]
        kernel.close(now=True)
        self._cached_kernel_properties = None

    def check_cached_kernel_spec(self, kernel_spec):
        """Test if kernel_spec corresponds to the cached kernel_spec."""
        if self._cached_kernel_properties is None:
            return False
        (
            cached_spec,
            cached_env,
            cached_argv,
            _,
        ) = self._cached_kernel_properties

        # Call interrupt_mode so the dict will be the same
        kernel_spec.interrupt_mode
        cached_spec.interrupt_mode

        if "PYTEST_CURRENT_TEST" in cached_env:
            # Make tests faster by using cached kernels
            # hopefully the kernel will never use PYTEST_CURRENT_TEST
            cached_env["PYTEST_CURRENT_TEST"] = (
                kernel_spec.env["PYTEST_CURRENT_TEST"])
        return (
            cached_spec.__dict__ == kernel_spec.__dict__
            and kernel_spec.argv == cached_argv
            and kernel_spec.env == cached_env
        )

    def get_cached_kernel(self, kernel_spec, cache=True):
        """Get a new kernel, and cache one for next time."""
        # Cache another kernel for next time.
        new_kernel_handler = KernelHandler.new_from_spec(kernel_spec)

        if not cache:
            # remove/don't use cache if requested
            self.close_cached_kernel()
            return new_kernel_handler

        # Check cached kernel has the same configuration as is being asked or
        # it crashed.
        cached_kernel_handler = None
        if self._cached_kernel_properties is not None:
            cached_kernel_handler = self._cached_kernel_properties[-1]
            if (
                not self.check_cached_kernel_spec(kernel_spec)
                or cached_kernel_handler._init_stderr
            ):
                # Close the kernel
                self.close_cached_kernel()
                cached_kernel_handler = None

        # Cache the new kernel
        self._cached_kernel_properties = (
            kernel_spec,
            kernel_spec.env,
            kernel_spec.argv,
            new_kernel_handler,
        )

        if cached_kernel_handler is None:
            return KernelHandler.new_from_spec(kernel_spec)

        return cached_kernel_handler
